# encoder.py
# taken from requests_toolbelt
# Apache license
# ff96b05243a3f5d4689cb776f4d03c9bda041957
#

# fuse.py
# ISC license
# https://github.com/akovari/fusepy/commit/847388cd52f060d5678980f20043afeedd166a55